"""
Module initialization file that exports database connectivity and utility functions.
"""
from .__about__ import __version__ as version
from .sql import sql
from .dbc.Utilities import (
    mask,
    get_host_ip,
    get_hostname,
)

# Public API
PUBLIC_API = [
    'sql',
    'mask',
    'version',
    'get_host_ip',
    'get_hostname',
]

__version__ = version
__all__ = PUBLIC_API



import json
import time
import inspect

from seSql.dbc.Utilities import mask
from seSql.dbc import ODBCError
from seSql.dbc.Exceptions import ODBCLoginFailed, ODBCInvalidOperation, ODBCDriverError

try:
    import pyodbc
    _pyodbc_available = True
except ImportError:  # pragma: no cover
    _pyodbc_available = False


validDrivers = ['ODBC Driver 18 for SQL Server', 'ODBC Driver 17 for SQL Server']


def odbcLoaded() -> bool:
    """
    Checks if the ODBC driver is loaded

    :return: bool
    """
    try:
        if not _pyodbc_available:  # pragma: no cover
            return False

        for validDriver in validDrivers:
            if validDriver in pyodbc.drivers():
                return True

        raise Exception()  # pragma: no cover

    except Exception:  # pragma: no cover
        return False


def odbcDriver() -> str:
    """
    Returns the ODBC driver
    :return: str
    """
    try:
        for validDriver in validDrivers:
            if validDriver in pyodbc.drivers():
                return validDriver

        raise Exception()  # pragma: no cover

    except Exception:  # pragma: no cover
        return "Error finding ODBC driver"


class odbcEngine:

    def __init__(self, server: str, port: int, user: str, password: str, trust: str, mars: str, db: str = "") -> None:
        """
        Initialize the odbcEngine object
        """
        self.__connStr = {
            "driver": odbcDriver(),
            "server": server,
            "port": port,
            "database": db,
            "user": user,
            "password": password,
            "trust": trust,
            "mars": mars
        }
        self.isConnected = False
        self.__cnxn = None
        self.connStats = None

    def connect(self):
        """
        Connect to database using pyodbc and ODBC Driver 17 or 18 for SQL Server
        """
        start_time = time.time()

        match self.__connStr["trust"]:
            case "yes":  # pragma: no cover
                connStr = f'DRIVER={{{self.__connStr["driver"]}}}; ' \
                          f'Server={self.__connStr["server"]},{self.__connStr["port"]}; ' \
                          f'DATABASE={self.__connStr["database"]}; ' \
                          f'UID={self.__connStr["user"]}; ' \
                          f'PWD={self.__connStr["password"]}; ' \
                          f'Trusted_Connection={self.__connStr["trust"]};'\
                          f'MARS_Connection={self.__connStr["mars"]};'
            case _:
                connStr = f'DRIVER={{{self.__connStr["driver"]}}}; ' \
                          f'Server={self.__connStr["server"]},{self.__connStr["port"]}; ' \
                          f'DATABASE={self.__connStr["database"]}; ' \
                          f'UID={self.__connStr["user"]}; ' \
                          f'PWD={self.__connStr["password"]};' \
                          f'MARS_Connection={self.__connStr["mars"]};'

        oJson = {
            f'odbc-{inspect.currentframe().f_code.co_name}': {
                'connection_ms': float(f'{0 * 1000:.2f}'),
                'connected': self.isConnected,
                'connStr': self.__connStr
            }}
        try:
            with pyodbc.connect(connStr, timeout=1, autocommit=False) as cnxn:
                self.__cnxn = cnxn
                self.isConnected = True
                oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['connected'] = self.isConnected

        except pyodbc.Error as ex:  # pragma: no cover
            oJson = {
                f'odbc-{inspect.currentframe().f_code.co_name}': {
                    'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                    'connected': self.isConnected,
                    'error': {}
                }
            }
            match f'{ODBCError.DefaultExceptionId(ex.args[1])}':  # pragma: no cover
                case "18456":
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['code'] = f'{ODBCError.DefaultExceptionId(ex.args[1])}'
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "ODBCLoginFailed"
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = f'{ODBCError.DefaultException(ex.args[1])}'
                    raise ODBCLoginFailed(json.dumps(oJson))
                case "851968" | "1048576":
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['code'] = f'{ODBCError.DefaultExceptionId(ex.args[1])}'
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "ODBCInvalidOperation"
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = f'{ODBCError.DefaultException(ex.args[1])}'
                    raise ODBCInvalidOperation(json.dumps(oJson))
                case _:
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['code'] = f'{ODBCError.DefaultExceptionId(ex.args[1])}'
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "ODBCDriverError"
                    oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = f'{ODBCError.DefaultException(ex.args[1])}'
                    raise ODBCDriverError(json.dumps(oJson))

        oJson = {
            f'odbc-{inspect.currentframe().f_code.co_name}': {
                'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                'connected': self.isConnected,
                'connStr': self.__connStr
            }}
        oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['connStr']['password'] = mask(oJson[f'odbc-{inspect.currentframe().f_code.co_name}']['connStr']['password'])
        # todo: verify if still needed
        # logger.info(json.dumps(oJson))
        self.connStats = json.dumps(oJson)

        return self.isConnected

    def query(self, query: str, stats: bool = False):
        if self.isConnected:
            start_time = time.time()
            try:
                with self.__cnxn.cursor() as cur:
                    cur.execute(query)
                    if cur.rowcount == -1:
                        r = [dict((cur.description[i][0], value)
                                  for i, value in enumerate(row)) for row in cur.fetchall()]
                        rCount = len(r)
                    else:
                        r = None
                        rCount = cur.rowcount
                    s = json.dumps({
                        f'odbc-{inspect.currentframe().f_code.co_name}': {
                            'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                            'rc': f'{cur.rowcount}:{rCount}',
                            'query': f'{query}'
                        }})

                if stats:
                    return {'results': r, 'stats': s}
                else:
                    return r

            except pyodbc.Error as e:  # pragma: no cover
                # todo: need to return
                oE = json.dumps({
                    f'odbc-{inspect.currentframe().f_code.co_name}': {
                        'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                        'query': f'{query}',
                        'error': f'{e}'
                    }})

                raise Exception(f'{oE}')

    @property
    def connectionString(self):  # pragma: no cover
        return self.__connStr



import re


class PyODBCError(Exception):
    """
    Handle errors for PyODBC. Offers an error message parser
    to apply specific logic depending on the error raise

    ODBC error identifier: 23000

    pyodbc_error_message (str) -- message raised by PyODBC
        Example:
            [23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] \
            Cannot insert explicit value for identity column in table \
            'building' when IDENTITY_INSERT is set to OFF.
            (544) (SQLExecDirectW) \
    """

    error_pattern = re.compile(
        # r"\[(?P<error_id>.*?)\] \[(?P<operator>.*?)\]\[(?P<driver>.*?)\]\[(?P<database_type>.*?)\](?P<error_message>.+?(?= \()) \((?P<sql_server_error_id>\d*?)\) \(SQLExecDirectW\)"
        r"\[(?P<error_id>.*?)] \[(?P<database_type>.*?)]\[(?P<driver>.*?)](?P<error_message>.+?(?= \()) \((?P<sql_server_error_id>\d*?)\) \((?P<sql_type>.*?)\)"
    )
    sql_error_code_pattern = re.compile(r"\((?P<sql_server_error_code>\d*?)\) \(SQLExecDirectW\)")
    column_pattern = re.compile(r"column \'(?P<column_name>.+?)\'")
    table_pattern = re.compile(r"table \'(?P<table_name>.+?)\'")
    pyodbc_error_code = 'HY000'

    def __init__(self, pyodbc_error_message: str) -> None:
        self._parse_error_message(pyodbc_error_message)

    def __str__(self) -> str:  # pragma: no cover
        return self.error_message

    def _parse_error_message(self, pyodbc_error_message: str) -> None:
        m = re.match(self.error_pattern, pyodbc_error_message)
        # self.operator = m.group('operator')
        if m:
            self.error_id = m.group('error_id')
            self.driver = m.group('driver')
            self.database_type = m.group('database_type')
            self.error_message = m.group('error_message')
            self.sql_server_error_id = m.group('sql_server_error_id')

    @classmethod
    def get_message(cls, pyodbc_exception: Exception) -> str:
        if pyodbc_exception.args[1] == cls.pyodbc_error_code:
            return pyodbc_exception.args[0]
        else:
            return pyodbc_exception.args[1]  # pragma: no cover

    @classmethod
    def get_pyodbc_code(cls, pyodbc_exception: Exception) -> str:
        if pyodbc_exception.args[1] == cls.pyodbc_error_code:
            return pyodbc_exception.args[1]
        else:
            return pyodbc_exception.args[0]  # pragma: no cover

    @staticmethod
    def get_exception(error_code: int):
        return {
            515: IdentityInsertNull,
            544: IdentityInsertSetToOff,
            2627: PrimaryKeyViolation,
            8114: FailedTypeConversion,
            102: IncorrectSyntax,
            32: InvalidNumberParametersSupplied
        }.get(error_code, DefaultException)

    @classmethod
    def get_sql_server_error_code(cls, pyodbc_code: str, message: str) -> int:
        """
        Parses error message raised by PyODBC and return SQL Server Error Code

        Looks for the following pattern:
            (544) (SQLExecDirectW) -> 544

        Args:
            pyodbc_error_message (str): Error string raised by PyODBC

        Returns:
            (int) - SQL Server Error Code
            :param message:
            :param pyodbc_code:
        """

        if pyodbc_code == cls.pyodbc_error_code:
            return 32
        else:  # pragma: no cover
            m = re.search(cls.sql_error_code_pattern, message)
            if m:
                return int(m.group('sql_server_error_code'))
            else:
                raise ValueError(f"Error raised is not from SQL Server: {message}")

    @classmethod
    def build_pyodbc_exception(cls, pyodbc_exception: Exception):
        pyodbc_code = cls.get_pyodbc_code(pyodbc_exception)
        error_message = cls.get_message(pyodbc_exception)
        error_code = cls.get_sql_server_error_code(pyodbc_code, error_message)
        exception = cls.get_exception(error_code)
        raise exception(error_message)


class IdentityInsertNull(PyODBCError):
    """
    Handle specific PyODBC error related to Null Value Inserted on Identity Column
    """

    def __init__(self, pyodbc_error_message):  # pragma: no cover
        super().__init__(pyodbc_error_message)
        m = re.search(self.table_pattern, self.error_message)
        self.table_name = m.group('table_name')
        m = re.search(self.column_pattern, self.error_message)
        self.column_name = m.group('column_name')


class IdentityInsertSetToOff(PyODBCError):
    """
    Handle specific PyODBC error related to Identity Not Set to On/Off
    """

    def __init__(self, pyodbc_error_message):
        super().__init__(pyodbc_error_message)
        m = re.search(self.table_pattern, self.error_message)
        self.table_name = m.group('table_name')


class FailedTypeConversion(PyODBCError):
    """
    Handle specific PyODBC error related to data type conversion
    """

    def __init__(self, pyodbc_error_message):
        super().__init__(pyodbc_error_message)


class PrimaryKeyViolation(PyODBCError):
    """
    Handle specific PyODBC error related to Primary Key Violation
    """

    def __init__(self, pyodbc_error_message):
        super().__init__(pyodbc_error_message)


class IncorrectSyntax(PyODBCError):
    """
    Handle specific PyODBC error related to incorrect syntax in a query
    """

    def __init__(self, pyodbc_error_message):
        super().__init__(pyodbc_error_message)


class DefaultExceptionId(PyODBCError):
    """
    Handle default PyODBC errors
    """

    def __init__(self, pyodbc_error_message):
        super().__init__(pyodbc_error_message)

    def __str__(self) -> str:
        return f"{self.sql_server_error_id}"


class DefaultException(PyODBCError):
    """
    Handle default PyODBC errors
    """

    def __init__(self, pyodbc_error_message):
        super().__init__(pyodbc_error_message)

    def __str__(self) -> str:
        try:
            return f"{self.error_message}"
        except AttributeError:
            return f"Error message not available"


class InvalidNumberParametersSupplied(Exception):
    def __init__(self, error_message) -> None:
        self.message = error_message

    def __str__(self) -> str:
        return self.message



from seSql.dbc.Exceptions import ODBCLoginFailed, JDBCSecureConnection, JDBCLoginFailed, JDBCConnectionTimeOut, JDBCSQLServerDriver, JDBCConnectionReset, ODBCInvalidOperation, ODBCDriverError, JDBCConnection, SQLConnectionException


def test_odbcLogin_failed():
    try:
        raise ODBCLoginFailed('ODBCLoginFailed')
    except ODBCLoginFailed as e:
        print(e)
        assert e.message == 'ODBCLoginFailed'


def test_odbcInvalidOperation():
    try:
        raise ODBCInvalidOperation('ODBCInvalidOperation')
    except ODBCInvalidOperation as e:
        print(e)
        assert e.message == 'ODBCInvalidOperation'


def test_odbcDriverError():
    try:
        raise ODBCDriverError('ODBCDriverError')
    except ODBCDriverError as e:
        print(e)
        assert e.message == 'ODBCDriverError'


def test_jdbcSecure_connection():
    try:
        raise JDBCSecureConnection('JDBCSecureConnection')
    except JDBCSecureConnection as e:
        print(e)
        assert e.message == 'JDBCSecureConnection'


def test_jdbcLogin_failed():
    try:
        raise JDBCLoginFailed('JDBCLoginFailed')
    except JDBCLoginFailed as e:
        print(e)
        assert e.message == 'JDBCLoginFailed'


def test_jdbcConnection_time_out():
    try:
        raise JDBCConnectionTimeOut('JDBCConnectionTimeOut')
    except JDBCConnectionTimeOut as e:
        print(e)
        assert e.message == 'JDBCConnectionTimeOut'


def test_jdbcSqlServer_driver():
    try:
        raise JDBCSQLServerDriver('JDBCSQLServerDriver')
    except JDBCSQLServerDriver as e:
        print(e)
        assert e.message == 'JDBCSQLServerDriver'


def test_jdbcConnectionReset():
    try:
        raise JDBCConnectionReset('JDBCConnectionReset')
    except JDBCConnectionReset as e:
        print(e)
        assert e.message == 'JDBCConnectionReset'


def test_jdbcConnection():
    try:
        raise JDBCConnection('JDBCConnection')
    except JDBCConnection as e:
        print(e)
        assert e.message == 'JDBCConnection'


def test_sqlConnectionException():
    try:
        raise SQLConnectionException('SQLConnectionException')
    except SQLConnectionException as e:
        print(e)
        assert e.message == 'SQLConnectionException'



import socket

# Constants for error messages
HOSTNAME_ERROR = "Unable to get Hostname"
IP_ERROR = "Unable to get IP"


def get_hostname() -> str:
    """
    Retrieves the current host name of the machine.

    Returns:
        str: The hostname of the machine or error message if unavailable
    """
    try:
        return socket.gethostname()
    except Exception:
        return HOSTNAME_ERROR


def get_host_ip() -> str:
    """
    Retrieves the IP address corresponding to the current host name.

    Returns:
        str: The IP address of the host or error message if unavailable
    """
    try:
        hostname = get_hostname()
        if hostname == HOSTNAME_ERROR:
            return IP_ERROR
        return socket.gethostbyname(hostname)
    except Exception:
        return IP_ERROR


def mask(to_mask: str) -> str:
    """
    Masks the given string
    Create a new string composed of '*' repeated (len(str1) - 5) times, followed by the last 5 characters of the original string

    :param to_mask:
    :return: str
    """
    return '*' * (len(to_mask) - 5) + to_mask[-5:]



import os.path
import pathlib
import time

import jaydebeapi
import inspect
import json
import glob

from seSql.dbc.Exceptions import JDBCConnectionReset, JDBCSecureConnection, JDBCLoginFailed, JDBCConnectionTimeOut, JDBCSQLServerDriver, JDBCConnection
from seSql.dbc.Utilities import mask

jdbcDrivers = {
    "mssql-12.2.0": {
        "classname": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
        "jar": "mssql-jdbc-12.2.0.jre8.jar"
    },
    "mssql-12.4.2": {
        "classname": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
        "jar": "mssql-jdbc-12.4.2.jre8.jar"
    },
    "mssql-0.0.0": {
        "classname": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
        "jar": "mssql-jdbc-0.0.0.jre8.jar"
    }
}

ROOT = pathlib.Path(__file__).parent.absolute()


def drivers() -> list:
    """
    Returns a list of loaded JDBC drivers

    :return: list
    """
    jarsPath = f'{ROOT}/jars'
    return [os.path.basename(files) for files in glob.glob(f'{jarsPath}/*')]


def jdbcLoaded(idx: int = -1) -> bool:
    """
    Checks if the JDBC driver is loaded

    :return: bool
    """
    try:
        return True if type([item for item in drivers()][idx]) is str else False
    except IndexError:
        return False


def jdbcDriver(idx: int = -1) -> str:
    """
    Returns the ODBC driver
    :return:
    """
    try:
        _dvr = [item for item in drivers()][idx] if jdbcLoaded() else "Error finding JDBC driver"
        return _dvr.replace("-jdbc-", "-").replace(".jre8.jar", "")
    except IndexError:
        return "Error finding JDBC driver"


def getDriverSettings():
    try:
        return jdbcDrivers[jdbcDriver()]
    except KeyError:  # pragma: no cover
        return jdbcDrivers["mssql-0.0.0"]


class jdbcEngine:
    def __init__(self, server: str, port: int, user: str, password: str, driver: str = 'mssql-12.4.2', encrypt: bool = True, trustServerCertificate: bool = True):
        """
        :rtype: object
        """
        self.__connStr = {
            "driver": driver,
            "server": f'jdbc:sqlserver://{server}',
            "port": port,
            "database": "",
            "user": user,
            "password": password,
            "encrypt": encrypt,
            "trustServerCertificate": trustServerCertificate
        }
        self.__cnxn = None
        self.isConnected = False
        self.connStats = None

    def connect(self) -> bool:
        driver = getDriverSettings()
        args = {
            "user": self.__connStr["user"],
            "password": self.__connStr["password"]
        }

        start_time = time.time()

        try:
            self.__cnxn = jaydebeapi.connect(
                driver['classname'],
                f'{self.__connStr["server"]}:{self.__connStr["port"]};encrypt={self.__connStr["encrypt"]};trustServerCertificate={self.__connStr["trustServerCertificate"]};',
                args,
                f"{ROOT}/jars/{driver['jar']}")
            self.isConnected = True

        except jaydebeapi.Error as e:  # pragma: no cover
            oJson = {
                f'jdbc-{inspect.currentframe().f_code.co_name}': {
                    'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                    'connected': self.isConnected,
                    'error': {
                        'exception': 'jayError',
                        'details': f'{e}'
                    }
                }
            }
            raise JDBCConnection(json.dumps(oJson))

        except Exception as e:  # pragma: no cover
            oJson = {
                f'jdbc-{inspect.currentframe().f_code.co_name}': {
                    'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                    'connected': self.isConnected,
                    'error': {}
                }
            }

            match e:  # pragma: no cover
                case _ if "driver could not establish a secure connection" in str(e).lower():
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "JDBCSecureConnection"
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = "Unable to establish a secure connection"
                    raise JDBCSecureConnection(json.dumps(oJson))

                case _ if "login failed" in str(e).lower():
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "JDBCLoginFailed"
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = "Login failed"
                    raise JDBCLoginFailed(json.dumps(oJson))

                case _ if "connect timed out" in str(e).lower():
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "JDBCConnectionTimeOut"
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = "Connect timed out. Verify the connection properties. Make sure that an instance of SQL Server is running on the host and accepting TCP/IP connections at the port"
                    raise JDBCConnectionTimeOut(json.dumps(oJson))

                case _ if "sqlserverdriver is not found" in str(e).lower():
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "JDBCSQLServerDriver"
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = "SQLServerDriver is not found"
                    raise JDBCSQLServerDriver(json.dumps(oJson))

                case _:
                    # logger.error(f'Exception: {e}')
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['exception'] = "JDBCConnectionReset"
                    oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['error']['details'] = "Make sure that TCP connections to the port are not blocked by a firewall. Verify the connection properties. Make sure that an instance of SQL Server is running on the host and accepting TCP/IP connections at the port."
                    raise JDBCConnectionReset(json.dumps(oJson))

        oJson = {
            f'jdbc-{inspect.currentframe().f_code.co_name}': {
                'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                'connected': self.isConnected,
                'connStr': self.__connStr
            }}

        oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['connStr']['password'] = mask(oJson[f'jdbc-{inspect.currentframe().f_code.co_name}']['connStr']['password'])
        # todo: verify if still needed
        # logger.info(json.dumps(oJson))

        self.connStats = json.dumps(oJson)

        return self.isConnected

    def query(self, query: str, stats: bool = False):
        if self.isConnected:
            start_time = time.time()
            try:
                with self.__cnxn.cursor() as curs:
                    curs.execute(query)
                    if curs.rowcount == -1:
                        r = [dict((curs.description[i][0], value)
                                  for i, value in enumerate(row)) for row in curs.fetchall()]
                        rCount = len(r)
                    else:
                        r = None
                        rCount = curs.rowcount

                    s = json.dumps({
                        f'jdbc-{inspect.currentframe().f_code.co_name}': {
                            'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                            'rc': f'{curs.rowcount}:{rCount}',
                            'query': f'{query}'
                        }})

                if stats:
                    return {'results': r, 'stats': s}
                else:
                    return r

            except Exception as e:
                # todo: verify if still needed
                oE = json.dumps({
                    f'jdbc-{inspect.currentframe().f_code.co_name}': {
                        'connection_ms': float(f'{(time.time() - start_time) * 1000:.2f}'),
                        'query': f'{query}',
                        'error': f'{e}'
                    }})

                raise Exception(f'{oE}')

    @property
    def connectionString(self):  # pragma: no cover
        return self.__connStr



import pytest
from seSql.dbc.ODBCError import PyODBCError, IdentityInsertNull, IdentityInsertSetToOff, PrimaryKeyViolation, \
    FailedTypeConversion, IncorrectSyntax, InvalidNumberParametersSupplied, DefaultExceptionId, DefaultException


def test_pyodbcerror_init():
    error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Cannot insert explicit value for " \
                    "identity column in table 'building' when IDENTITY_INSERT is set to OFF. (544) (SQLExecDirectW) "
    exception = PyODBCError(error_message)
    assert exception.error_id == "23000"
    assert exception.driver == "ODBC Driver 17 for SQL Server"
    assert exception.database_type == "Microsoft"
    assert exception.error_message == "[SQL Server] Cannot insert explicit value for identity column in table 'building' when IDENTITY_INSERT is set to OFF."
    assert exception.sql_server_error_id == "544"


# def test_identity_insert_null_init():
#     error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Cannot insert explicit value for " \
#                     "identity column in table 'employees' when IDENTITY_INSERT is set to OFF. (544) (SQLExecDirectW) "
#     exception = IdentityInsertNull(error_message)
#     assert exception.table_name == "employees"
#     assert exception.column_name is None


def test_identity_insert_set_to_off_init():
    error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Cannot insert explicit value for " \
                    "identity column in table 'buildings' when IDENTITY_INSERT is set to OFF. (544) (SQLExecDirectW) "
    exception = IdentityInsertSetToOff(error_message)
    assert exception.table_name == "buildings"


def test_primary_key_violation_init():
    error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Violation of PRIMARY KEY constraint " \
                    "'PK_table_name'. Cannot insert duplicate key in object 'dbo.table_name'. The duplicate key value is '(value)'. (2627) (SQLExecDirectW)"
    exception = PrimaryKeyViolation(error_message)
    assert isinstance(exception, PrimaryKeyViolation)


def test_failed_type_conversion_init():
    error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Error converting data type varchar to int. (8114) (SQLExecDirectW)"
    exception = FailedTypeConversion(error_message)
    assert isinstance(exception, FailedTypeConversion)


def test_incorrect_syntax_init():
    error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Incorrect syntax near 'some_syntax'. (102) (SQLExecDirectW)"
    exception = IncorrectSyntax(error_message)
    assert isinstance(exception, IncorrectSyntax)


def test_invalid_number_parameters_supplied():
    error_message = "Invalid number of parameters supplied."
    exception = InvalidNumberParametersSupplied(error_message)
    assert exception.message == error_message


def test_pyodbc_error_build_pyodbc_exception():
    pyodbc_exception = Exception("Some message", PyODBCError.pyodbc_error_code)
    with pytest.raises(InvalidNumberParametersSupplied):
        PyODBCError.build_pyodbc_exception(pyodbc_exception)


def test_pyodbc_error_get_sql_server_error_code():
    message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Cannot insert explicit value for " \
              "identity column in table 'customers' when IDENTITY_INSERT is set to OFF. (544) (SQLExecDirectW) "
    pyodbc_code = "HY000"
    error_code = PyODBCError.get_sql_server_error_code(pyodbc_code, message)
    assert error_code == 32


def test_defaultexceptionid_str():
    error_message = "[23000] [Microsoft][ODBC Driver 17 for SQL Server][SQL Server] Cannot insert explicit value for " \
                    "identity column in table 'orders' when IDENTITY_INSERT is set to OFF. (544) (SQLExecDirectW) "
    exception = DefaultExceptionId(error_message)
    assert str(exception) == "544"


def test_invalid_number_parameters_supplied_message():
    error_message = "Invalid number of parameters supplied to the query."
    error = InvalidNumberParametersSupplied(error_message)
    assert str(error) == error_message


def test_invalid_number_parameters_supplied_empty_message():
    error_message = ""
    error = InvalidNumberParametersSupplied(error_message)
    assert str(error) == error_message


def test_default_exception_str_representation():
    """Test the string representation of DefaultException."""
    error_message = "Test default exception error message."
    exception = DefaultException(error_message)
    assert str(exception) != error_message


def test_default_exception_inheritance():
    """Test that DefaultException is a subclass of Exception."""
    error_message = "Another test message."
    exception = DefaultException(error_message)
    assert isinstance(exception, Exception)


import json

from typing import Optional

from .__about__ import __version__

from .dbc.Exceptions import SQLConnectionException, SQLQueryException
from .dbc.JDBC import jdbcLoaded, jdbcDriver, jdbcEngine
from .dbc.ODBC import odbcLoaded, odbcDriver, odbcEngine


class sql:
    def __init__(self):
        self.__cnxn = None
        self.__isConnected = False
        self.__driverOverride = None
        self.__jdbcDriver = None
        self.__jdbcLoaded = None
        self.__odbcDriver = None
        self.__odbcLoaded = None
        self.__connStr = None
        self.__connStatus = None
        self.__connStats = None

    def connect(self,
                server: Optional[str] = None,
                port: Optional[int] = None,
                user: Optional[str] = None,
                password: Optional[str] = None,
                trust: Optional[str] = None,
                db: Optional[str] = "",
                trustServerCertificate: Optional[bool] = True,
                mars: Optional[str] = 'no',
                driverOverride: Optional[str] = None,
                ):

        self.__odbcLoaded = odbcLoaded()
        self.__odbcDriver = odbcDriver()
        self.__jdbcLoaded = jdbcLoaded()
        self.__jdbcDriver = jdbcDriver()
        self.__driverOverride = driverOverride.lower() if driverOverride.lower() in ["odbc", "jdbc"] else "odbc"

        self.__connStr = {
            "server": server,
            "port": port,
            "user": user,
            "password": password,
            "trust": trust,
            "db": db,
            "trustServerCertificate": trustServerCertificate,
            "mars": mars,
            "driverOverride": driverOverride
        }

        if self.__driverOverride == "odbc" and self.__odbcLoaded:
            self.__driverOverride = "odbc"
        elif self.__driverOverride == "jdbc" and self.__jdbcLoaded:
            self.__driverOverride = "jdbc"
        else:  # pragma: no cover
            self.__driverOverride = "jdbc"

        match self.__driverOverride:
            case "odbc":
                self.__cnxn = odbcEngine(
                    server=self.__connStr['server'],
                    port=self.__connStr['port'],
                    db=self.__connStr['db'],
                    user=self.__connStr['user'],
                    password=self.__connStr['password'],
                    trust=self.__connStr['trust'],
                    mars=self.__connStr['mars'],
                )
                self._logging()

                try:
                    self.__cnxn.connect()
                    self.__isConnected = self.__cnxn.isConnected
                    self.__connStats = self.__cnxn.connStats
                except Exception as e:  # pragma: no cover
                    raise SQLConnectionException(f'{e}')

            case _:
                self.__cnxn = jdbcEngine(
                    server=self.__connStr['server'],
                    port=self.__connStr['port'],
                    user=self.__connStr['user'],
                    password=self.__connStr['password'],
                    trustServerCertificate=self.__connStr['trustServerCertificate']
                )
                self._logging()

                try:
                    self.__cnxn.connect()
                    self.__isConnected = self.__cnxn.isConnected
                    self.__connStats = self.__cnxn.connStats
                except Exception as e:  # pragma: no cover
                    raise SQLConnectionException(f'{e}')

    def query(self,
              query: str,
              stats: bool = False
              ):
        try:
            return self.__cnxn.query(query, stats=stats)
        except Exception as e:
            raise SQLQueryException(f'{e}')

    def _logging(self):
        oSqlDriver = {
            "driver-info":
                {
                    "using": self.__driverOverride,
                    "odbc": {
                        "loaded": self.__odbcLoaded,
                        "driver": self.__odbcDriver
                    },
                    "jdbc": {
                        "loaded": self.__jdbcLoaded,
                        "driver": self.__jdbcDriver
                    }
                }
        }
        self.__connStatus = json.dumps(oSqlDriver)

    @property
    def isConnected(self) -> bool:
        return self.__isConnected
        # return True

    @property
    def ConnectedStatus(self):

        return self.__connStatus

    @property
    def ConnectedStats(self):

        return self.__connStats

    @property
    def version(self):
        return __version__



class JDBCSecureConnection(Exception):
    """Exception raised for "encrypt" property is set to "true" and "trustServerCertificate" property is set to "false" but the driver could not establish a secure connection to SQL Server by using Secure Sockets Layer (SSL) encryption."""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class JDBCConnection(Exception):
    """Exception raised for jaydebeapi.Error."""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class JDBCLoginFailed(Exception):
    """Exception raised for Login failed for user/pw"""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class JDBCConnectionTimeOut(Exception):
    """Exception raised for Connect timed out. Verify the connection properties. Make sure that an instance of SQL Server is running on the host and accepting TCP/IP connections at the port"""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class JDBCConnectionReset(Exception):
    """Make sure that TCP connections to the port are not blocked by a firewall. Verify the connection properties. Make sure that an instance of SQL Server is running on the host and accepting TCP/IP connections at the port."""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class JDBCSQLServerDriver(Exception):
    """Exception raised for SQLServerDriver is not found."""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class ODBCLoginFailed(Exception):
    """Exception raised for Login failed for user/pw"""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class ODBCInvalidOperation(Exception):
    """Exception raised for Trust setting"""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class ODBCDriverError(Exception):
    """Exception raised for anything not caught by ODBC driver"""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class SQLQueryException(Exception):
    """Exception raised for anything caught by an SQL query."""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)


class SQLConnectionException(Exception):
    """Exception raised for anything caught by an SQL query."""
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return str(self.message)



import json

from seCore.CustomLogging import logger
from seSql import sql

# Constants
DB_CONFIG = {
    "server": "SQL5101.site4now.net",
    "port": 1433,
    "user": "db_a82904_cybernetic_admin",
    "password": "tb7qiwqer8mee68",
    "trust": "no",
    "driverOverride": "jdbc",
    "mars": "no"
}


def test_version():
    oSql = sql()
    oSql.connect(**DB_CONFIG)

    assert isinstance(oSql.version, str), "Version must be a string"


def test_jdbc():
    oSql = sql()
    oSql.connect(**DB_CONFIG)

    json.dumps({
        'seSql': {
            'driver': json.loads(oSql.ConnectedStatus)["driver-info"],
        }
    })
    json.dumps({
        'seSql': {
            'ep': 'connect',
            'stats': json.loads(oSql.ConnectedStats),
        }
    })

    if oSql.isConnected:
        # oSql.query("SELECT @@version as version")
        oSql.query("select @@version as version")

        # oSql.query("select * from dbo.winequality_red;")

        try:
            oSql.query("select * from dbo.winequality_red;")
        except Exception as e:
            logger.error(f'Exception: {e}')

        try:
            oSql.query("select * from dba.winequality_red;")
        except Exception as e:
            logger.error(f'Exception: {e}')

        try:
            oResponse = oSql.query("select @@version as version;", stats=True)
            logger.info(oResponse['stats'])
            logger.info(oResponse['results'])
        except Exception as e:
            logger.error(f'{e}')

        try:
            oResponse = oSql.query("delete from dbo.seSql_Settings where value like '%dev';")
            logger.info(oResponse)
        except Exception as e:
            logger.error(f'{e}')



import socket

from seSql.dbc.Utilities import get_host_ip, get_hostname

# Mock data for testing
HOSTNAME_ERROR = "Unable to get Hostname"
IP_ERROR = "Unable to get IP"
TEST_HOSTNAME = "test_hostname"
TEST_IP = "192.168.1.1"


def test_get_hostname_success(monkeypatch):
    """Test the get_hostname function when successfully retrieving the hostname."""

    def mock_gethostname():
        return TEST_HOSTNAME

    monkeypatch.setattr(socket, "gethostname", mock_gethostname)
    result = get_hostname()
    assert result == TEST_HOSTNAME


def test_get_hostname_failure(monkeypatch):
    """Test the get_hostname function when an exception occurs."""

    def mock_gethostname():
        raise Exception("Error retrieving hostname")

    monkeypatch.setattr(socket, "gethostname", mock_gethostname)
    result = get_hostname()
    assert result == HOSTNAME_ERROR


def test_get_hostname_error(monkeypatch):
    """Test the get_hostname function when retrieval fails."""

    def mock_gethostname():
        raise OSError("Unable to retrieve hostname")

    monkeypatch.setattr(socket, "gethostname", mock_gethostname)

    result = get_hostname()
    assert result == HOSTNAME_ERROR


def test_get_host_ip_success(monkeypatch):
    """Test the get_host_ip function when successful."""

    def mock_gethostname():
        return TEST_HOSTNAME

    def mock_gethostbyname(hostname):
        if hostname == TEST_HOSTNAME:
            return TEST_IP
        raise socket.gaierror("Unknown host")

    monkeypatch.setattr(socket, "gethostname", mock_gethostname)
    monkeypatch.setattr(socket, "gethostbyname", mock_gethostbyname)

    result = get_host_ip()
    assert result == TEST_IP


def test_get_host_ip_hostname_error(monkeypatch):
    """Test get_host_ip function when hostname retrieval fails."""

    def mock_gethostname():
        return HOSTNAME_ERROR

    monkeypatch.setattr(socket, "gethostname", mock_gethostname)

    result = get_host_ip()
    assert result == IP_ERROR


def test_get_host_ip_socket_error(monkeypatch):
    """Test get_host_ip function when socket.gethostbyname raises an error."""

    def mock_gethostname():
        return TEST_HOSTNAME

    def mock_gethostbyname(hostname):
        raise socket.gaierror("Unknown host")

    monkeypatch.setattr(socket, "gethostname", mock_gethostname)
    monkeypatch.setattr(socket, "gethostbyname", mock_gethostbyname)

    result = get_host_ip()
    assert result == IP_ERROR
